'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var FS = _interopDefault(require('fs'));
var Path = _interopDefault(require('path'));
var vueVirtualTextdocument = require('@vuedx/vue-virtual-textdocument');
var vscodeUri = require('vscode-uri');

const PATCHED_METHODS = Symbol('Vue Patched Methods');
function tryPatchMethod(target, methodName, createOverride) {
    const patched = target[PATCHED_METHODS] || [];
    if (patched.includes(methodName))
        return;
    let fn = target[methodName];
    if (typeof fn === 'function') {
        try {
            fn = fn.bind(target);
        }
        catch {
            // - ignore
        }
    }
    target[methodName] = createOverride(fn);
    patched.push(methodName);
    target[PATCHED_METHODS] = patched;
}

function removeVirtualSuffixFromFileName(fileName) {
    if (vueVirtualTextdocument.isVirtualFile(String(fileName))) {
        return fileName.substr(0, fileName.lastIndexOf(vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR));
    }
    return fileName;
}
const virtualFileRegex = new RegExp(`(?<=\\.vue)${vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR}(script|template|style|customBlock|render)(__[0-9]+)?(\\.[A-Za-z0-9_-]+)?`, 'g');
function removeVirtualSuffixFromText(text) {
    return text.replace(virtualFileRegex, '');
}
function prepareDocumentSpan(span) {
    span.fileName = removeVirtualSuffixFromFileName(span.fileName);
    if (span.originalFileName) {
        span.originalFileName = removeVirtualSuffixFromFileName(span.originalFileName);
    }
}
function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}

class ProxyDocumentStore extends vueVirtualTextdocument.DocumentStore {
    get(fileNameOrUri) {
        return super.get(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    has(fileNameOrUri) {
        return super.has(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    set(fileNameOrUri, document) {
        return super.set(vueVirtualTextdocument.asUri(fileNameOrUri), document);
    }
    delete(fileNameOrUri) {
        return super.delete(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
}
class PluginContext {
    constructor(typescript) {
        this.typescript = typescript;
        this.watchers = new Map();
        this.store = new ProxyDocumentStore((uri) => {
            const fileName = vscodeUri.URI.parse(uri).fsPath;
            const content = this.typescript.sys.readFile(fileName) || '';
            return vueVirtualTextdocument.VueTextDocument.create(uri, 'vue', 0, content, 'javascript');
        }, (uri) => {
            const fileName = vscodeUri.URI.parse(uri).fsPath;
            return vscodeUri.URI.file(this.projectService.toPath(fileName)).toString();
        });
    }
    get serviceHost() {
        return this._serverHost;
    }
    get projectService() {
        return this._projectService;
    }
    log(message) {
        if (this.projectService) {
            this.projectService.logger.info(`Vue.js:: ${message}`);
        }
    }
    error(message) {
        if (this.projectService) {
            this.projectService.logger.msg(`Vue.js:: ${message} ${message.stack}`, this.typescript.server.Msg.Err);
        }
    }
    getSupportedVirtualDocumentFileNames(document) {
        const fileNames = [document.getBlockDocumentFileName('script')];
        if (document.descriptor.template?.content) {
            fileNames.push(document.getBlockDocumentFileName('render'));
        }
        return fileNames;
    }
    watchVirtualFile(fileName, callback) {
        if (!this.watchers.has(fileName)) {
            this.watchers.set(fileName, []);
        }
        this.watchers.get(fileName)?.push(callback);
    }
    triggerVirtualFileWatchers(fileName, event) {
        this.watchers.get(fileName)?.forEach((callback) => {
            try {
                callback(fileName, event);
            }
            catch (error) {
                this.error(error);
            }
        });
    }
    stopVirtualFileWatcher(fileName, callback) {
        if (!this.watchers.has(fileName))
            return;
        const watchers = this.watchers.get(fileName);
        watchers.splice(watchers.indexOf(callback), 1);
    }
    tryCreateScriptInfo(fileName) {
        try {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                this.store.get(vueVirtualTextdocument.getContainingFile(fileName)); // Trigger VueTextDocument creation.
                patchScriptInfo(this, this.projectService.getOrCreateScriptInfoForNormalizedPath(this.typescript.server.toNormalizedPath(vueVirtualTextdocument.getContainingFile(fileName)), false));
            }
        }
        catch (error) {
            this.log(`[ERROR] Cannot create ScriptInfo for "${fileName}". Reason: ${error.message} ${error.stack}`);
        }
    }
    load(info) {
        this.log(`Loading Vue plugin: ${info.project.getProjectName()}`);
        this._serverHost = info.serverHost;
        this._projectService = info.project.projectService;
        patchProjectService(this);
        patchServiceHost(this);
        patchLanguageServiceHost(this, info.languageServiceHost);
    }
}
function patchProjectService(context) {
    patchExtraFileExtensions(context);
}
function patchExtraFileExtensions(context) {
    const extraFileExtensions = [
        { extension: 'vue', isMixedContent: false, scriptKind: context.typescript.ScriptKind.Deferred },
    ];
    tryPatchMethod(context.projectService, 'setHostConfiguration', (setHostConfiguration) => {
        context.log(`[patch] Add support for vue extension. (ProjectService)`);
        return (args) => {
            context.log(`[------------->] setting extraFileExtensions`);
            const current = context.projectService.hostConfiguration
                .extraFileExtensions;
            if (args.extraFileExtensions) {
                args.extraFileExtensions.push(...extraFileExtensions);
            }
            else if (!current || !current.some((ext) => ext.extension === 'vue')) {
                args.extraFileExtensions = [...extraFileExtensions];
            }
            return setHostConfiguration(args);
        };
    });
    if (context.projectService.hostConfiguration.extraFileExtensions?.some((ext) => ext.extension === 'vue')) {
        return;
    }
    context.projectService.setHostConfiguration({ extraFileExtensions: [] });
}
function patchLanguageServiceHost(context, languageServiceHost) {
    patchGetScriptFileNames(context, languageServiceHost);
    patchModuleResolution(context, languageServiceHost);
}
function patchGetScriptFileNames(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'getScriptFileNames', (getScriptFileNames) => {
        context.log(`[patch] Override getScriptFileNames to expand .vue files to virtual files. (LanguageServerHost)`);
        const previousVueFiles = new Set();
        return () => {
            const fileNames = [];
            [...getScriptFileNames(), ...previousVueFiles].forEach((fileName) => {
                if (vueVirtualTextdocument.isVueFile(fileName)) {
                    const document = context.store.get(fileName);
                    if (document) {
                        if (!previousVueFiles.has(fileName))
                            previousVueFiles.add(fileName);
                        context.getSupportedVirtualDocumentFileNames(document).forEach((fileName) => {
                            context.tryCreateScriptInfo(fileName);
                            fileNames.push(fileName);
                        });
                    }
                    else {
                        previousVueFiles.delete(fileName);
                    }
                }
                else {
                    fileNames.push(fileName);
                }
            });
            return fileNames;
        };
    });
}
function patchFileExists(context) {
    tryPatchMethod(context.serviceHost, 'fileExists', (fileExists = context.typescript.sys.fileExists) => {
        context.log(`[patch] Override fileExists to check containing file for virtual files. (ServiceHost)`);
        return (fileName) => {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                const vueFileName = vueVirtualTextdocument.getContainingFile(fileName);
                if (fileExists(vueFileName)) {
                    const document = context.store.get(vueFileName);
                    if (!document)
                        return false;
                    const { selector } = vueVirtualTextdocument.parseVirtualFileName(fileName);
                    return document.getBlockDocumentFileName(selector) === fileName;
                }
                return false;
            }
            return fileExists(fileName);
        };
    });
}
function patchReadFile(context) {
    tryPatchMethod(context.serviceHost, 'readFile', (readFile) => {
        context.log(`[patch] Override readFile to check containing file for virtual files. (ServiceHost)`);
        return (fileName) => {
            if (vueVirtualTextdocument.isVirtualFile(fileName)) {
                context.log(`host.readFile("${fileName}")`);
                return context.store.get(vueVirtualTextdocument.getContainingFile(fileName))?.getBlockDocument(fileName)?.getText();
            }
            return readFile ? readFile(fileName) : context.typescript.sys.readFile(fileName);
        };
    });
}
function patchModuleResolution(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'resolveModuleNames', (resolveModuleNames) => {
        context.log(`[patch] Override resolveModuleNames to resolve imports from .vue files. (LanguageServerHost)`);
        return (moduleNames, containingFile, reusedNames, redirectedReferences, options) => {
            if (vueVirtualTextdocument.isVueFile(containingFile))
                throw new Error('A .vue file should not be part of TS program.');
            const newModuleNames = moduleNames.map((moduleName) => vueVirtualTextdocument.isVueFile(moduleName) ? moduleName + vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + 'script' : moduleName);
            // TODO: Support paths mapped to .vue files, if needed.
            const result = resolveModuleNames
                ? resolveModuleNames(newModuleNames, containingFile, reusedNames, redirectedReferences, options)
                : [];
            result.map((resolved) => {
                if (resolved && vueVirtualTextdocument.isVirtualFile(resolved.resolvedFileName)) {
                    context.tryCreateScriptInfo(resolved.resolvedFileName);
                }
            });
            return result;
        };
    });
}
function patchServiceHost(context) {
    patchWatchFile(context);
    patchFileExists(context);
    patchReadFile(context);
}
function patchWatchFile(context) {
    tryPatchMethod(context.serviceHost, 'watchFile', (watchFile) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return (fileName, callback, pollingInterval, options) => {
            const shouldPatchWatcher = vueVirtualTextdocument.isVirtualFile(fileName);
            if (shouldPatchWatcher) {
                context.watchVirtualFile(fileName, callback);
                const originalFileName = fileName;
                const originalCallback = callback;
                fileName = vueVirtualTextdocument.getContainingFile(fileName);
                callback = (_, eventKind) => originalCallback(originalFileName, eventKind);
            }
            const watcher = watchFile(fileName, callback, pollingInterval, options);
            if (shouldPatchWatcher) {
                return {
                    close() {
                        watcher.close();
                        context.stopVirtualFileWatcher(fileName, callback);
                    },
                };
            }
            return watcher;
        };
    });
}
function patchScriptInfo(context, scriptInfo) {
    if (!scriptInfo)
        throw new Error('ScriptInfo is required.');
    tryPatchMethod(scriptInfo, 'editContent', (editContent) => {
        context.log(`[patch] Override editContent() of "${scriptInfo.fileName}" to sync virtual files. (ScriptInfo)`);
        return (start, end, newText) => {
            const document = context.store.get(scriptInfo.fileName);
            if (!document)
                throw new Error('VueTextDocument should exist for every ScriptInfo.');
            const range = { start: document.positionAt(start), end: document.positionAt(end) };
            editContent(start, end, newText);
            const preVirtuals = context.getSupportedVirtualDocumentFileNames(document);
            vueVirtualTextdocument.VueTextDocument.update(document, [{ range, text: newText }], getLastNumberFromVersion(scriptInfo.getLatestVersion()));
            const curVirtuals = context.getSupportedVirtualDocumentFileNames(document);
            curVirtuals.forEach((fileName) => {
                context.triggerVirtualFileWatchers(fileName, context.typescript.FileWatcherEventKind.Changed);
            });
            preVirtuals
                .filter((fileName) => !curVirtuals.includes(fileName))
                .forEach((fileName) => {
                context.triggerVirtualFileWatchers(fileName, context.typescript.FileWatcherEventKind.Deleted);
            });
        };
    });
}

function prepareCodeFixAction(fileName, result) {
    result.forEach(item => {
        item.changes.forEach(item => {
            // TODO: Can be new file!!!
            item.fileName = removeVirtualSuffixFromFileName(fileName);
        });
    });
    return result;
}

function prepareCallHierarchy(fileName, result) {
    if (Array.isArray(result)) {
        result.forEach(item => {
            item.file = removeVirtualSuffixFromFileName(item.file);
        });
    }
    else {
        result.file = removeVirtualSuffixFromFileName(result.file);
    }
    return result;
}
function prepareCallHierachyIncomingCalls(fileName, result) {
    result.forEach(item => {
        item.from.file = removeVirtualSuffixFromFileName(item.from.file);
    });
    return result;
}
function prepareCallHierachyOutgoingCalls(fileName, result) {
    result.forEach(item => {
        item.to.file = removeVirtualSuffixFromFileName(item.to.file);
    });
    return result;
}

function prepareClassifiedSpans(fileName, result) {
    return result;
}
function prepareClassifications(fileName, result) {
    return result;
}

function prepareCompletionsInfo(result) {
    return result;
}
function remapCompletionsInfo(result, document) {
    if (document) {
        result.entries.forEach((entry) => {
            if (entry.replacementSpan) {
                const position = document.getSourceOffsetAt(entry.replacementSpan.start);
                if (position != null) {
                    entry.replacementSpan.start = position;
                }
                else {
                    entry.replacementSpan = undefined;
                }
            }
        });
        return result;
    }
}
function prepareCompletionsEntryDetail(result) {
    result.codeActions?.forEach((codeAction) => {
        codeAction.changes?.forEach((change) => {
            change.fileName = removeVirtualSuffixFromFileName(change.fileName);
        });
    });
    result.displayParts.forEach((displayPart) => (displayPart.text = removeVirtualSuffixFromText(displayPart.text)));
    return result;
}
function remapCompletionsEntryDetail(result, document) {
    if (document) {
        result.codeActions?.forEach((codeAction) => {
            codeAction.changes?.forEach((change) => {
                change.textChanges = change.textChanges.filter((textChange) => {
                    const offset = document.getSourceOffsetAt(textChange.span.start);
                    if (offset != null) {
                        textChange.span.start = offset;
                        return true;
                    }
                    return false;
                });
            });
        });
        return result;
    }
}
function prepareSymbol(result) {
    return result;
}

function prepareDefinitionAndBoundSpan(fileName, result) {
    result.definitions?.forEach(prepareDocumentSpan);
    return result;
}
function prepareDefinitionInfo(fileName, result) {
    result.forEach(prepareDocumentSpan);
    return result;
}
function prepareImplementationLocation(fileName, result) {
    result.forEach(prepareDocumentSpan);
    return result;
}
function prepareReferenceEntry(fileName, result) {
    result.forEach(prepareDocumentSpan);
    return result;
}
function prepareReferenceSymbol(fileName, result) {
    result.forEach((symbol) => {
        prepareDocumentSpan(symbol.definition);
        symbol.references.forEach(prepareDocumentSpan);
    });
    return result;
}

function prepareSyntacticDiagnostics(result) {
    result.forEach(prepareDiagnostic);
    return result;
}
function remapSyntacticDiagnostics(result, document) {
    if (document) {
        return result.filter(remapDiagnostic.bind(null, document));
    }
    return [];
}
function prepareSuggestionDiagnostics(result) {
    result.forEach(prepareDiagnostic);
    return result;
}
function remapSuggestionDiagnostics(result, document) {
    if (document) {
        return result.filter(remapDiagnostic.bind(null, document));
    }
    return [];
}
function prepareSemanticDiagnostics(result) {
    result.forEach(prepareDiagnostic);
    return result;
}
function remapSemanticDiagnosts(result, document) {
    if (document) {
        return result.filter(remapDiagnostic.bind(null, document));
    }
    return [];
}
function prepareDiagnostic(diagnostic) {
    if (typeof diagnostic.messageText === 'string') {
        diagnostic.messageText = removeVirtualSuffixFromText(diagnostic.messageText);
    }
}
function remapDiagnostic(document, diagnostic) {
    if (diagnostic.start != null) {
        let start = document.getSourceOffsetAt(diagnostic.start);
        let end = diagnostic.length ? document.getSourceOffsetAt(diagnostic.start + diagnostic.length - 1) : start;
        if (start == null || end == null) {
            return true;
        }
        else if (diagnostic.length) {
            end += 1;
        }
        diagnostic.start = start;
        diagnostic.length = end - start;
    }
    return true;
}

function prepareApplicableRefactorInfo(fileName, result) {
    return result;
}
function prepareRefactorEditInfo(fileName, result) {
    if (result.renameFilename) {
        result.renameFilename = removeVirtualSuffixFromFileName(result.renameFilename);
    }
    result.edits.forEach((edit) => {
        edit.fileName = removeVirtualSuffixFromFileName(edit.fileName);
    });
    return result;
}
function remapTextSpan(span, document) {
    if (span.start != null) {
        let start = document.getSourceOffsetAt(span.start);
        let end = span.length ? document.getSourceOffsetAt(span.start + span.length - 1) : start;
        if (start == null || end == null) {
            return false;
        }
        else if (span.length) {
            end += 1;
        }
        span.start = start;
        span.length = end - start;
    }
    return true;
}

function prepareDocumentHighlights(fileName, result) {
    result.forEach((highlight) => {
        highlight.fileName = removeVirtualSuffixFromFileName(highlight.fileName);
        highlight.highlightSpans.forEach((span) => {
            if (span.fileName)
                span.fileName = removeVirtualSuffixFromFileName(span.fileName);
        });
    });
    return result;
}

function prepareQuickInfo(result) {
    result.displayParts?.forEach((part) => {
        part.text = removeVirtualSuffixFromText(part.text);
    });
    return result;
}
function remapQuickInfo(result, document) {
    if (document) {
        const offset = document.getSourceOffsetAt(result.textSpan.start);
        if (offset != null) {
            result.textSpan.start = offset;
            return result;
        }
    }
}

function prepareRenameInfo(fileName, result) {
    if ('fileToRename' in result && result.fileToRename) {
        result.fileToRename = removeVirtualSuffixFromFileName(result.fileToRename);
    }
    return result;
}

function prepareSignatureHelpItems(fileName, result) {
    return result;
}

function isNumber(any) {
    return typeof any === 'number';
}
function flat(items) {
    if (typeof Array.prototype.flat === 'function') {
        return items.flat(Infinity);
    }
    const result = [];
    items.forEach((item) => {
        if (Array.isArray(item)) {
            result.push(...flat(item));
        }
        else {
            result.push(item);
        }
    });
    return result;
}
function firstOrDefault(items, defaultValue) {
    if (items.length)
        return items[0];
    return defaultValue();
}
function chain(value) {
    return {
        next(fn) {
            if (value == undefined) {
                return chain(undefined);
            }
            try {
                return chain(fn(value));
            }
            catch (error) {
                console.log('Error in chain: ' + error.message + ' ' + error.stack);
                return chain(undefined);
            }
        },
        // @ts-ignore
        end(defaultValue) {
            return value == undefined ? defaultValue : value;
        },
    };
}
class VirtualFileSystemHelper {
    constructor(documents, logger) {
        this.documents = documents;
        this.logger = logger;
    }
    getContainingFile(fileName) {
        return vueVirtualTextdocument.getContainingFile(fileName);
    }
    isVueFile(fileName) {
        return vueVirtualTextdocument.isVueFile(fileName);
    }
    isVirtualFile(fileName) {
        return vueVirtualTextdocument.isVirtualFile(fileName);
    }
    isRenderFunction(fileName) {
        return vueVirtualTextdocument.isVirtualFile(fileName) && fileName.includes(vueVirtualTextdocument.VIRTUAL_FILENAME_SEPARATOR + 'render');
    }
    getVueDocument(fileName) {
        return this.documents.get(fileName);
    }
    getVirtualDocument(fileName) {
        return this.getVueDocument(this.getContainingFile(fileName))?.getBlockDocument(fileName);
    }
    getRenderFunctionDocument(fileName) {
        return this.getVueDocument(this.getContainingFile(fileName))?.getBlockDocument('render');
    }
    getFileNameAt(fileName, position, isRenderFunctionAllowed = false) {
        if (this.isVueFile(fileName)) {
            let virtualFileName = undefined;
            const document = this.documents.get(fileName);
            if (!document)
                return;
            const block = document.blockAt(position);
            if (!block || !block.content)
                return;
            else if (block.type === 'script') {
                virtualFileName = document.getBlockDocumentFileName('script');
            }
            else if (block.type === 'template' && block.content && isRenderFunctionAllowed) {
                virtualFileName = document.getBlockDocumentFileName('render');
            }
            return virtualFileName;
        }
        return fileName;
    }
    getInnerFileNames(fileName, isRenderFunctionAllowed = false) {
        if (this.isVueFile(fileName)) {
            const document = this.documents.get(fileName);
            if (!document)
                return [];
            const fileNames = [];
            if (document.descriptor.script?.content) {
                fileNames.push(document.getBlockDocumentFileName('script'));
            }
            if (isRenderFunctionAllowed && document.descriptor.template?.content) {
                fileNames.push(document.getBlockDocumentFileName('render'));
            }
            return fileNames;
        }
        return [fileName];
    }
}
const VUE_LANGUAGE_SERVER = Symbol('Vue Language Server');
class VueLanguageServer {
    constructor(context) {
        this.context = context;
        this.fileSystem = new VirtualFileSystemHelper(context.store, context);
    }
    getPositionInGeneratedSource(fileName, position, length = 1) {
        if (this.fileSystem.isRenderFunction(fileName)) {
            const document = this.fileSystem.getRenderFunctionDocument(fileName);
            if (document)
                return document.getGeneratedOffsetAt(position, length);
        }
        return position;
    }
    getProgram() {
        const program = this.service.getProgram();
        if (!program) {
            this.context.log('Unxpected undefined program.');
            return this.prevProgram;
        }
        this.prevProgram = program;
        return program;
    }
    getSyntacticDiagnostics(fileName) {
        const diagnostics = flat(this.fileSystem.getInnerFileNames(fileName, true).map((fileName) => {
            const diagnostics = this.service.getSyntacticDiagnostics(fileName);
            if (this.fileSystem.isRenderFunction(fileName)) {
                return remapSyntacticDiagnostics(diagnostics, this.fileSystem.getRenderFunctionDocument(fileName));
            }
            return diagnostics;
        }));
        return prepareSyntacticDiagnostics(diagnostics);
    }
    getSemanticDiagnostics(fileName) {
        const diagnostics = flat(this.fileSystem.getInnerFileNames(fileName, true).map((fileName) => {
            const diagnostics = this.service.getSemanticDiagnostics(fileName);
            if (this.fileSystem.isRenderFunction(fileName)) {
                this.context.log(`${this.context.projectService.getScriptInfo(fileName)?.getLatestVersion()} has ${diagnostics.length} => ${fileName}`);
                return remapSemanticDiagnosts(diagnostics, this.fileSystem.getRenderFunctionDocument(fileName));
            }
            return diagnostics;
        }));
        if (this.fileSystem.isVueFile(fileName)) {
            this.context.log(`${this.context.projectService.getScriptInfo(fileName)?.getLatestVersion()} => ${fileName}`);
        }
        return prepareSemanticDiagnostics(diagnostics);
    }
    getSuggestionDiagnostics(fileName) {
        const diagnostics = flat(this.fileSystem.getInnerFileNames(fileName, true).map((fileName) => {
            const diagnostics = this.service.getSuggestionDiagnostics(fileName);
            if (this.fileSystem.isRenderFunction(fileName)) {
                return remapSuggestionDiagnostics(diagnostics, this.fileSystem.getRenderFunctionDocument(fileName));
            }
            return diagnostics;
        }));
        return prepareSuggestionDiagnostics(diagnostics);
    }
    getSyntacticClassifications(fileName, span) {
        return chain(this.fileSystem.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getSyntacticClassifications(fileName, span))
            .next((classifications) => prepareClassifiedSpans(fileName, classifications))
            .end([]);
    }
    getSemanticClassifications(fileName, span) {
        return chain(this.fileSystem.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getSemanticClassifications(fileName, span))
            .next((classifications) => prepareClassifiedSpans(fileName, classifications))
            .end([]);
    }
    getEncodedSyntacticClassifications(fileName, span) {
        return chain(this.fileSystem.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getEncodedSyntacticClassifications(fileName, span))
            .next((classifications) => prepareClassifications(fileName, classifications))
            .end({ spans: [], endOfLineState: this.context.typescript.EndOfLineState.None });
    }
    getEncodedSemanticClassifications(fileName, span) {
        return chain(this.fileSystem.getFileNameAt(fileName, span.start))
            .next((fileName) => this.service.getEncodedSemanticClassifications(fileName, span))
            .next((classifications) => prepareClassifications(fileName, classifications))
            .end({ spans: [], endOfLineState: this.context.typescript.EndOfLineState.None });
    }
    getCompletionsAtPosition(fileName, position, options) {
        return chain(this.fileSystem.getFileNameAt(fileName, position, true))
            .next((fileName) => chain(this.getPositionInGeneratedSource(fileName, position))
            .next((position) => this.service.getCompletionsAtPosition(fileName, position, options))
            .next((result) => this.fileSystem.isRenderFunction(fileName)
            ? remapCompletionsInfo(result, this.fileSystem.getRenderFunctionDocument(fileName))
            : result)
            .end())
            .next(prepareCompletionsInfo)
            .end();
    }
    getCompletionEntryDetails(fileName, position, name, formatOptions, source, preferences) {
        return chain(this.fileSystem.getFileNameAt(fileName, position, true))
            .next((fileName) => chain(this.getPositionInGeneratedSource(fileName, position))
            .next((position) => this.service.getCompletionEntryDetails(fileName, position, name, formatOptions, source, preferences))
            .next((result) => this.fileSystem.isRenderFunction(fileName)
            ? remapCompletionsEntryDetail(result, this.fileSystem.getRenderFunctionDocument(fileName))
            : result)
            .end())
            .next(prepareCompletionsEntryDetail)
            .end();
    }
    getCompletionEntrySymbol(fileName, position, name, source) {
        return chain(this.fileSystem.getFileNameAt(fileName, position, true))
            .next((fileName) => chain(this.getPositionInGeneratedSource(fileName, position))
            .next((position) => this.getCompletionEntrySymbol(fileName, position, name, source))
            .end())
            .next(prepareSymbol)
            .end();
    }
    getQuickInfoAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position, true))
            .next((fileName) => chain(this.getPositionInGeneratedSource(fileName, position))
            .next((position) => this.service.getQuickInfoAtPosition(fileName, position))
            .next((result) => this.fileSystem.isRenderFunction(fileName)
            ? remapQuickInfo(result, this.fileSystem.getRenderFunctionDocument(fileName))
            : result)
            .end())
            .next(prepareQuickInfo)
            .end();
    }
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return chain(this.fileSystem.getFileNameAt(fileName, startPos))
            .next((fileName) => this.service.getNameOrDottedNameSpan(fileName, startPos, endPos))
            .end();
    }
    getBreakpointStatementAtPosition(fileName, position) {
        // NOTE: Breakpoints in template cannot be supported as we are using a fake render function.
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getBreakpointStatementAtPosition(fileName, position))
            .end();
    }
    getSignatureHelpItems(fileName, position, options) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getSignatureHelpItems(fileName, position, options))
            .next((items) => prepareSignatureHelpItems(fileName, items))
            .end();
    }
    getRenameInfo(fileName, position, options) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getRenameInfo(fileName, position, options))
            .next((info) => {
            this.context.log(`server.getRenameInfo() =>  ${JSON.stringify(info, null, 2)}`);
            return info;
        })
            .next((info) => prepareRenameInfo(fileName, info))
            .end({ canRename: false, localizedErrorMessage: '' });
    }
    findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename))
            .next((locations) => {
            this.context.log(`server.findRenameLocations() =>  ${JSON.stringify(locations, null, 2)}`);
            return locations;
        })
            .next((locations) => locations.filter((location) => {
            const fileName = location.fileName;
            location.fileName = removeVirtualSuffixFromFileName(location.fileName);
            if (location.originalFileName) {
                location.originalFileName = removeVirtualSuffixFromFileName(location.originalFileName);
            }
            if (this.fileSystem.isRenderFunction(fileName)) {
                return remapTextSpan(location.textSpan, this.fileSystem.getRenderFunctionDocument(fileName));
            }
            return true;
        }))
            .end();
    }
    getSmartSelectionRange(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getSmartSelectionRange(fileName, position))
            .end({ textSpan: { start: position, length: 0 } });
    }
    getDefinitionAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDefinitionAtPosition(fileName, position))
            .next((result) => prepareDefinitionInfo(fileName, result))
            .end();
    }
    getDefinitionAndBoundSpan(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDefinitionAndBoundSpan(fileName, position))
            .next((result) => prepareDefinitionAndBoundSpan(fileName, result))
            .end();
    }
    getTypeDefinitionAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getTypeDefinitionAtPosition(fileName, position))
            .next((result) => prepareDefinitionInfo(fileName, result))
            .end();
    }
    getImplementationAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getImplementationAtPosition(fileName, position))
            .next((result) => prepareImplementationLocation(fileName, result))
            .end();
    }
    getReferencesAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getReferencesAtPosition(fileName, position))
            .next((result) => prepareReferenceEntry(fileName, result))
            .end();
    }
    findReferences(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.findReferences(fileName, position))
            .next((result) => prepareReferenceSymbol(fileName, result))
            .end();
    }
    getDocumentHighlights(fileName, position, filesToSearch) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDocumentHighlights(fileName, position, filesToSearch
            .map((fileName) => this.fileSystem.getInnerFileNames(fileName))
            .filter(Boolean)
            .flat()))
            .next((highlights) => prepareDocumentHighlights(fileName, highlights))
            .end();
    }
    getOccurrencesAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getOccurrencesAtPosition(fileName, position))
            .next((references) => prepareReferenceEntry(fileName, references))
            .end();
    }
    getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles) {
        return this.service.getNavigateToItems(searchValue, maxResultCount, fileName, // TODO: Convert .vue file to script
        excludeDtsFiles);
    }
    getNavigationBarItems(fileName) {
        return flat(this.fileSystem.getInnerFileNames(fileName).map((fileName) => this.service.getNavigationBarItems(fileName)));
    }
    getNavigationTree(fileName) {
        // TODO: Do we need to merge navigation tree from <script> and <template> blocks?
        return firstOrDefault(this.fileSystem.getInnerFileNames(fileName).map((fileName) => this.service.getNavigationTree(fileName)), () => ({
            kind: this.context.typescript.ScriptElementKind.moduleElement,
            text: '<unknown>',
            kindModifiers: 'none',
            nameSpan: undefined,
            spans: [],
        }));
    }
    prepareCallHierarchy(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.prepareCallHierarchy(fileName, position))
            .next((result) => prepareCallHierarchy(fileName, result))
            .end();
    }
    provideCallHierarchyIncomingCalls(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.provideCallHierarchyIncomingCalls(fileName, position))
            .next((result) => prepareCallHierachyIncomingCalls(fileName, result))
            .end([]);
    }
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.provideCallHierarchyOutgoingCalls(fileName, position))
            .next((result) => prepareCallHierachyOutgoingCalls(fileName, result))
            .end([]);
    }
    getOutliningSpans(fileName) {
        return flat(this.fileSystem.getInnerFileNames(fileName).map((fileName) => this.service.getOutliningSpans(fileName)));
    }
    getTodoComments(fileName, descriptors) {
        return flat(this.fileSystem.getInnerFileNames(fileName).map((fileName) => this.service.getTodoComments(fileName, descriptors)));
    }
    getBraceMatchingAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getBraceMatchingAtPosition(fileName, position))
            .end([]);
    }
    getIndentationAtPosition(fileName, position, options) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getIndentationAtPosition(fileName, position, options))
            .end(0);
    }
    getFormattingEditsForRange(fileName, start, end, options) {
        return chain(this.fileSystem.getFileNameAt(fileName, start))
            .next((fileName) => this.service.getFormattingEditsForRange(fileName, start, end, options))
            .end([]);
    }
    getFormattingEditsForDocument(fileName, options) {
        return flat(this.fileSystem
            .getInnerFileNames(fileName)
            .map((fileName) => this.service.getFormattingEditsForDocument(fileName, options)));
    }
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getFormattingEditsAfterKeystroke(fileName, position, key, options))
            .end([]);
    }
    getDocCommentTemplateAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getDocCommentTemplateAtPosition(fileName, position))
            .end();
    }
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.isValidBraceCompletionAtPosition(fileName, position, openingBrace))
            .end(false);
    }
    getJsxClosingTagAtPosition(fileName, position) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getJsxClosingTagAtPosition(fileName, position))
            .end();
    }
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return chain(this.fileSystem.getFileNameAt(fileName, position))
            .next((fileName) => this.service.getSpanOfEnclosingComment(fileName, position, onlyMultiLine))
            .end();
    }
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return chain(this.fileSystem.getFileNameAt(fileName, start))
            .next((fileName) => this.service.getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences))
            .next((fixes) => prepareCodeFixAction(fileName, fixes))
            .end([]);
    }
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return this.service.getCombinedCodeFix(scope, fixId, formatOptions, preferences);
    }
    applyCodeActionCommand(action, formatSettings) {
        return this.service.applyCodeActionCommand(action, formatSettings);
    }
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return chain(this.fileSystem.getFileNameAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos))
            .next((fileName) => this.service.getApplicableRefactors(fileName, positionOrRange, preferences))
            .next((refactors) => prepareApplicableRefactorInfo(fileName, refactors))
            .end([]);
    }
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return chain(this.fileSystem.getFileNameAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos))
            .next((fileName) => this.service.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences))
            .next((refactor) => prepareRefactorEditInfo(fileName, refactor))
            .end();
    }
    organizeImports(scope, formatOptions, preferences) {
        return chain(scope)
            .next((scope) => {
            if (scope.type === 'file') {
                const fileName = this.fileSystem.getInnerFileNames(scope.fileName).find(Boolean);
                if (fileName)
                    return { ...scope, fileName };
                return undefined; // abort
            }
            return scope;
        })
            .next((scope) => this.service.organizeImports(scope, formatOptions, preferences))
            .next((results) => {
            results.forEach((result) => {
                if (this.fileSystem.isVirtualFile(result.fileName)) {
                    // TODO: Ensure offset padding for script blocks
                    result.fileName = removeVirtualSuffixFromFileName(result.fileName);
                }
            });
            return results;
        })
            .end([]);
    }
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        // TODO: Handle rename.
        return this.service.getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences);
    }
    decorate(languageService) {
        if (VUE_LANGUAGE_SERVER in languageService) {
            return languageService;
        }
        this.context.log(`[patch] Decorating TypeScript language server.`);
        const proxy = Object.create(null);
        proxy[VUE_LANGUAGE_SERVER] = true;
        this.service = languageService;
        for (const key of Object.keys(languageService)) {
            if (key in languageService) {
                proxy[key] =
                    languageService[key] && key in this
                        ? (...args) => {
                            const result = this[key].apply(this, args);
                            return result;
                        }
                        : languageService[key];
            }
        }
        return proxy;
    }
}

let context;
let server;
function init({ typescript }) {
    if (!context) {
        context = new PluginContext(typescript);
        server = new VueLanguageServer(context);
    }
    return {
        create(info) {
            const version = findVueVersionIn(info.project.getCurrentDirectory());
            if (!isVue3(version)) {
                return info.languageService;
            }
            else {
                context.load(info);
                return server.decorate(info.languageService);
            }
        },
    };
}
function isVue3(version) {
    if (!version)
        return true; // assume Vue 3 by default.
    return version.startsWith('3.');
}
function findVueVersionIn(directory) {
    const packageFile = findNearestPackageFile(directory);
    if (packageFile) {
        const vuePackageFile = Path.resolve(Path.dirname(packageFile), 'node_modules', 'vue', 'package.json');
        if (FS.existsSync(vuePackageFile)) {
            return require(vuePackageFile).version;
        }
        // dependencies not installed.
        const pkg = require(packageFile);
        const version = pkg.dependencies?.vue || pkg.devDependencies?.vue;
        if (!version)
            return;
        const justNumbers = version.replace(/[^0-9.]/, '');
        if (justNumbers.startsWith('3.'))
            return justNumbers;
        if (version === 'next')
            return '3.0.0';
        return '2.x';
    }
}
function findNearestPackageFile(directory) {
    if (!directory)
        return;
    const packageFile = Path.resolve(directory, 'package.json');
    if (FS.existsSync(packageFile))
        return packageFile;
    if (directory === '/')
        return;
    return findNearestPackageFile(Path.dirname(directory));
}

module.exports = init;
//# sourceMappingURL=typescript-plugin-vue.js.map
